from vcCommand import *
import re
import vcMatrix
import vcProductTypeFilter


USE_VCID_IF_POSSIBLE = False


app = getApplication()
cmd = getCommand()
sim = getSimulation()
process_ctr = sim.ProcessController
flow_group_mgr = process_ctr.FlowGroupManager
product_mgr = process_ctr.ProductTypeManager
pattern_mgr = process_ctr.AssemblyPatternManager


def OnStart():
  global source_prop, similar_prop, use_existing_prop, create_prop
  
  source_prop = confirm_property(cmd, VC_STRING, 'Source', 'Component', ['Component', 'Node', 'Feature'])
  similar_prop = confirm_property(cmd, VC_BOOLEAN, 'Find Similar Products By Name', True)
  use_existing_prop = confirm_property(cmd, VC_BOOLEAN, 'Use Existing Products', True)
  create_prop = confirm_property(cmd, VC_BUTTON, 'Create Assembly')
  
  executeInActionPanel()
  
  similar_prop.OnChanged = format_ui
  create_prop.OnChanged = create_assembly


def format_ui(*arg):
  use_existing_prop.IsVisible = similar_prop.Value


def confirm_property(object, type, name, init_value = None, step_values = None, min_value = None, max_value = None, is_visible = True, rebuild = True, update = False):
  prop = object.getProperty(name)
  if not prop:
    constraint = VC_PROPERTY_DEFAULT
    if step_values:
      constraint = VC_PROPERTY_STEP
    elif min_value and max_value:
      constraint = VC_PROPERTY_LIMIT
    prop = object.createProperty(type, name, constraint)
    if step_values:
      prop.StepValues = step_values
    elif min_value and max_value:
      prop.MinValue = min_value
      prop.MaxValue = max_value
    if init_value:
      prop.Value = init_value
    prop.IsVisible = is_visible
    prop.ChangeOnRebuild = rebuild
    prop.ChangeOnUpdate = update
  return prop


def create_assembly(*args):
  source_comp = None
  sel = app.SelectionManager.getSelection(VC_SELECTION_COMPONENT)
  if not sel:
    print 'No source component selected.'
    return
  source_comp = sel[0]
  if source_prop.Value == 'Component':
    create_assembly_from_components(source_comp)
  elif source_prop.Value == 'Node':
    create_components_from_nodes(source_comp)
    create_assembly_from_components(source_comp)
  else: #if source_prop.Value == 'Feature':
    create_components_from_features(source_comp)
    create_assembly_from_components(source_comp)
  print 'Assembly type created.'


def create_assembly_from_components(source_comp, find_similar_products_by_name = True):
  check_name(source_comp)
  flow_group = create_new_flow_group(source_comp)
  assembly_type = create_assembly_type(source_comp, flow_group)
  assembly_steps = {}
  products = {}
  comps = [source_comp]
  while comps:
    comp = comps.pop(0)
    check_name(comp)
    children = comp.ComponentChildren
    children.sort(key = lambda x: x.Name)
    comps.extend(children)
    if similar_prop.Value and use_existing_prop.Value:
      product_name = get_base_name(comp.Name)
      product_type =  product_mgr.findProductType(product_name)
      if not product_type:
        product_type = create_product_type(comp, flow_group)
    else:
      if similar_prop.Value:
        product_name = get_base_name(comp.Name)
      else:
        product_name = comp.Name
      if not product_name in products.keys():
        product_type = create_product_type(comp, flow_group)
    products[product_name] = product_type
    parent_step = None
    if comp != source_comp and comp.Parent.Component.Name in assembly_steps.keys():
      parent_step = assembly_steps[comp.Parent.Component.Name]
    step = create_assembly_step(assembly_type, comp, product_type, parent_step)
    assembly_steps[comp.Name] = step


def create_components_from_nodes(source_comp):
  comps = {}
  nodes = [source_comp]
  while nodes:
    node = nodes.pop(0)
    children = node.Children
    children.sort(key = lambda x: x.Name)
    nodes.extend(children)
    if node == source_comp:
      comps[node.Name] = source_comp
      continue
    comp_name = node.Name
    comp = app.findComponent(comp_name)
    if comp:
      for i in range(2,1000):
        comp_name = '%s #%i' % (node.Name, (i))
        comp = app.findComponent(comp_name)
        if not comp:
          break
    comp = app.createComponent()
    comp.Name = comp_name
    comps[node.Name] = comp
    parent_comp = comps[node.Parent.Name]
    parent_comp.attach(comp)
    comp.PositionMatrix = node.PositionMatrix
    for feat in node.RootFeature.Children:
      comp.RootFeature.attach(feat)
  
  for node in comp.Children:
    node.delete()
  
  for comp in comps.values():
    comp.rebuild()
  
  app.render()


def create_components_from_features(source_comp):
  source_feat = None
  for feat in source_comp.RootFeature.Children:
    if not feat.Type in [VC_FRAME]:
      source_feat = feat
      break
  if not source_feat:
    print 'No features on selected component.'
    return
  if len(source_comp.RootFeature.Children) > 1:
    print 'Warning: Selected component has more than 1 root level features. Building assembly on %s.' % (source_feat.Name)
  
  comps = {}
  feats = [source_feat]
  while feats:
    feat = feats.pop(0)
    children = feat.Children
    children.sort(key = lambda x: x.Name)
    feats.extend(children)
    if feat == source_feat:
      comps[feat.Name] = source_comp
      continue
    comp_name = feat.Name
    comp = app.findComponent(comp_name)
    if comp:
      for i in range(2,1000):
        comp_name = '%s #%i' % (feat.Name, (i))
        comp = app.findComponent(comp_name)
        if not comp:
          break
    comp = app.createComponent()
    comp.Name = comp_name
    comps[feat.Name] = comp
    parent_comp = comps[feat.Parent.Name]
    parent_comp.attach(comp)
    comp.PositionMatrix = feat.PositionMatrix
    comp.RootFeature.attach(feat)
    feat.PositionMatrix = vcMatrix.new()
  
  for comp in comps.values():
    comp.rebuild()
  
  app.render()


def create_new_flow_group(source_comp):
  base_name = get_base_name(source_comp.Name)
  group = flow_group_mgr.findGroup(base_name)
  if not group:
    flow_group_name = base_name
  else:
    for i in range(2,1000):
      flow_group_name = '%s #%i' % (base_name, (i))
      group = flow_group_mgr.findGroup(flow_group_name)
      if not group:
        break
  flow_group = flow_group_mgr.createGroup(flow_group_name)
  return flow_group


def create_product_type(comp, flow_group):
  base_name = get_base_name(comp.Name)
  product_type = product_mgr.findProductType(base_name)
  if not product_type:
    product_name = base_name
  else:
    for i in range(2, 1000):
      product_name = '%s #%i' % (base_name, (i))
      product_type = product_mgr.findProductType(product_name)
      if not product_type:
        break
  
  children = comp.Children
  for child in children:
    sim.World.attach(child)
  
  product_type = product_mgr.createProductType(product_name, flow_group, False)
  if USE_VCID_IF_POSSIBLE and comp.VCID:
    product_type.ComponentUri = 'VCID:%s' % (comp.VCID)
  else:
    product_type.ComponentUri = 'component:%s' % (comp.Name)
  
  for prop in comp.Properties:
    if prop.Name in ['Name', 'Visible', 'BackfaceMode', 'PDFExportLevel', 'BOM', 'BOMname', 'BOMdescription', 'Category', 'SimulationLevel']:
      continue
    product_prop = product_type.createComponentProperty(prop.Type, prop.Name)
    product_prop.Value = prop.Value
  
  for child in children:
    comp.attach(child)
  
  return product_type


def create_assembly_type(comp, flow_group):
  base_name = get_base_name(comp.Name) + '_Assembly'
  assembly_type = product_mgr.findProductType(base_name)
  if not assembly_type:
    assembly_name = base_name
  else:
    for i in range(2, 1000):
      assembly_name = '%s #%i' % (base_name, (i))
      assembly_type = product_mgr.findProductType(assembly_name)
      if not assembly_type:
        break
  assembly_type = product_mgr.createProductType(assembly_name, flow_group, True)
  return assembly_type


def create_assembly_step(assembly_type, comp, product_type, parent_step = None):
  if not parent_step:
    step = assembly_type.RootStep
  else:
    step = assembly_type.createStepTo(parent_step, -1)
  step.Name = comp.Name
  pattern_name = step.Name
  pattern = pattern_mgr.findPattern(pattern_name)
  if pattern:
    for i in range(2,1000):
      pattern_name = '%s #%i' % (step.Name, (i))
      pattern = pattern_mgr.findPattern(pattern_name)
      if not pattern:
        break
  pattern = pattern_mgr.createPattern(pattern_name)
  step.Pattern = pattern
  slot = pattern.createSlot(0)
  if parent_step:
    step.RelativeFrame = comp.PositionMatrix
  item = slot.Item
  item.DefaultProductType = product_type
  try:
    # 4.7 or newer
    filter = item.setFilter(VC_FILTER_CONFIGURABLEPRODUCTTYPE)
    filter.AcceptAllProductTypes = False
    filter.AcceptedProductTypes = [product_type]
  except:
    # 4.6 or older
    item.AcceptAllProductTypes = False
    item.AcceptedProductTypes = [product_type]
  return step


def get_base_name(name):
  exp = '\s?(#|_)[0-9]+$'
  name = name
  match = re.search(exp, name)
  if match:
    name = name[:match.start()]
  return name


def check_name(comp):
  chars = '+'
  for char in chars:
    comp.Name = comp.Name.replace(char, '_')